/******************************************************************************
*
* project name:    SNOW
* initial date:    26/06/2000
* authors:         thomas.nussbaumer@gmx.net
*                  xdanger1@caramail.com
* description:     snow effect demonstration
*
* !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
* NOTE: current implementation uses TIGCC 0.93
* !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
*
* for TI89 compile with:   tigcc -O3 -DUSE_TI89  dissolve.c
* for TI92P compile with:  tigcc -O3 -DUSE_TI92P dissolve.c
*
*
* -----------------------------------------------------------------------------
* Revision History
* -----------------------------------------------------------------------------
*
* v1.20: 18/02/2002
*        small changes (update for TIGCC 0.93)    [xdanger]
*        Different versions for 89 and 92+        [xdanger]
*        size reduced using static function trick [tnussb]
*
* v1.00: 26/06/2000
*        initial version
*
*******************************************************************************/

#if defined(USE_TI89) && defined(USE_TI92P)
#error cannot produce both versions at once
#elif defined(USE_TI89)
#define C89_92(x,y) (x)
#elif defined(USE_TI92P)
#define C89_92(x,y) (y)
#else
#error Please specify either -DUSE_TI89 or -DUSE_TI92P on commandline
#endif

#define OPTIMIZE_ROM_CALLS  // for optimized size
#define NO_EXIT_SUPPORT     // we don't need the exit support -> smaller code

#include <tigcclib.h>


/*---------------------------------------------------------------------------*/
/* some constants                                                            */
/*---------------------------------------------------------------------------*/

#define NR_PARTICLES  500      // number of max. snow particles to handle
#define RAND_RANGE    C89_92(160*98,240*126) // range in which pixels are selected

#define FREE_ENTRY    0xffff   // indicates an unused particle entry

/*---------------------------------------------------------------------------*/
/* structure which stores a snow particle's position                         */
/*---------------------------------------------------------------------------*/
typedef struct {
    unsigned short offset;  // byte offset of particle (if this equals to
                            // FREE_ENTRY than this entry is free to re-use)
    unsigned char  mask;    // bit mask of pixel
    // 1 byte wasted...
} particle;


/*---------------------------------------------------------------------------*/
/* global variables                                                          */
/*---------------------------------------------------------------------------*/
particle* positions = NULL;  // array of particle positions
short     count     = 0;     // number of snow particles


/*===========================================================================*/
/* fills particles array                                                     */
/*                                                                           */
/* !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! */
/*                                                                           */
/* NOTE: an important optimization hint is to define internal routines which */
/*       are used just once or twice static.                                 */
/*       This enables the optimizer to completely inline them !!!!           */
/*===========================================================================*/
static void FillParticlesArray(void) {
    short          i;
    short          j=0;
    unsigned short pix;
    unsigned short offset;
    unsigned char  mask;

    for (i=count;i<NR_PARTICLES;i++) {
        //----------------------------------------------
        // search for next free entry in particles array
        //----------------------------------------------
        while (j<NR_PARTICLES && positions[j].offset != FREE_ENTRY) j++;
        if (j>=NR_PARTICLES) return;

        //----------------------------------------------
        // select a random pixel and calculate its byte
        // offset and its bitmask
        //----------------------------------------------
        pix = random(RAND_RANGE);
        offset = pix >> 3;
        mask   = 0x80 >> (pix % 8);

        //----------------------------------------------
        // if the randomly selected pixel is set, store
        // its values into the actual particles array
        // entry
        //----------------------------------------------
        if (*((unsigned char *)LCD_MEM+offset) & mask) {
            count++;
            positions[j].offset = offset;
            positions[j].mask   = mask;
        }
    }
}


/*===========================================================================*/
/* process particles in particles array                                      */
/*                                                                           */
/* !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! */
/*                                                                           */
/* NOTE: an important optimization hint is to define internal routines which */
/*       are used just once or twice static.                                 */
/*       This enables the optimizer to completely inline them !!!!           */
/*===========================================================================*/
static void ProcessParticlesArray(void) {
    short i;

    for (i=0;i<NR_PARTICLES;i++) {
        //--------------------------------------------------
        // if this entry is not marked as free, check if the
        // pixel just below it is set. If there is no pixel
        // just below it, remove the actual pixel and move
        // it one line down
        //--------------------------------------------------
        if (positions[i].offset != FREE_ENTRY) {
            if (!(*((unsigned char *)LCD_MEM + positions[i].offset + 30) & positions[i].mask)) {
                *((unsigned char *)LCD_MEM + positions[i].offset) &= ~positions[i].mask;
                positions[i].offset += 30;
                *((unsigned char *)LCD_MEM + positions[i].offset) |= positions[i].mask;
                if (positions[i].offset / 30 >= 127) {
                    positions[i].offset = FREE_ENTRY;
                    count--;
                }
            }
            else {
              //-------------------------------------------
              // if there is a pixel exactly one line below
              // the actual one, mark the actual particles
              // array entry as free, because it cannot
              // "fall" down
              //-------------------------------------------
              positions[i].offset = FREE_ENTRY;
              count--;
            }
        }
    }
}


/*===========================================================================*/
/* main routine: where all the fun starts ...                                */
/*===========================================================================*/
void _main(void) {
    short          i;
    short          key;
    LCD_BUFFER     lcd;
    register void *kbq = kbd_queue();

    //------------------------------------------------
    // try to allocate memory for particles
    //------------------------------------------------
    if (!(positions = (particle*)malloc(sizeof(particle)*NR_PARTICLES))) {
        ST_helpMsg("ERROR: not enough memory");
        return;
    }

    LCD_save(lcd);

    //--------------------------------------------------
    // using a label in combination with a goto is not
    // really good style, but keeps the size of the file
    // low ;-)
    //--------------------------------------------------
begin:
    ClrScr();
    randomize();
    //------------------------------------------------
    // mark all entries in the particles array as free
    // and initialize the global variables
    //------------------------------------------------
    for (i=0;i<NR_PARTICLES;i++) positions[i].offset = FREE_ENTRY;

    count = 0;

#ifdef USE_TI89
    // asm block which draws black line on top of screen
    // (well, some inline ASM magic - but not really hard to guess whats going on)
asm("pea (%a0)\n"
    "move.l %d0,-(%sp)\n"
    "moveq #-1,%d0\n"
    "lea 0x579a.w,%a0\n"
    "move.l %d0,(%a0)+\n"
    "move.l %d0,(%a0)+\n"
    "move.l %d0,(%a0)+\n"
    "move.l %d0,(%a0)+\n"
    "move.l %d0,(%a0)\n"
    "move.l (%sp)+,%d0\n"
    "move.l (%sp)+,%a0");
#else
    // asm block which draws black line on top of screen
    // (well, some inline ASM magic - but not really hard to guess whats going on)
asm("pea    (%a0)\n"
    "move.l %d0,-(%sp)\n"
    "moveq  #-1,%d0\n"
    "lea    0x5ae2.w,%a0\n"
    "move.l %d0,(%a0)+\n"
    "move.l %d0,(%a0)+\n"
    "move.l %d0,(%a0)+\n"
    "move.l %d0,(%a0)+\n"
    "move.l %d0,(%a0)+\n"
    "move.l %d0,(%a0)+\n"
    "move.l %d0,(%a0)+\n"
    "move.w %d0,(%a0)\n"
    "move.l (%sp)+,%d0\n"
    "move.l (%sp)+,%a0");
#endif

    //---------------------------------------------
    // main loop
    //---------------------------------------------
    while (1) {
#ifdef USE_TI89
    // asm block which draws black line on top of screen
    // (well, some inline ASM magic - but not really hard to guess whats going on)
asm("pea    (%a0)\n"
    "move.l %d0,-(%sp)\n"
    "moveq  #-1,%d0\n"
    "lea    0x4c00.w,%a0\n"
    "move.l %d0,(%a0)+\n"
    "move.l %d0,(%a0)+\n"
    "move.l %d0,(%a0)+\n"
    "move.l %d0,(%a0)+\n"
    "move.l %d0,(%a0)\n"
    "move.l (%sp)+,%d0\n"
    "move.l (%sp)+,%a0");
#else
    // asm block which draws black line on top of screen
    // (well, some inline ASM magic - but not really hard to guess whats going on)
asm("pea    (%a0)\n"
    "move.l %d0,-(%sp)\n"
    "moveq  #-1,%d0\n"
    "lea    0x4c00.w,%a0\n"
    "move.l %d0,(%a0)+\n"
    "move.l %d0,(%a0)+\n"
    "move.l %d0,(%a0)+\n"
    "move.l %d0,(%a0)+\n"
    "move.l %d0,(%a0)+\n"
    "move.l %d0,(%a0)+\n"
    "move.l %d0,(%a0)+\n"
    "move.w %d0,(%a0)\n"
    "move.l (%sp)+,%d0\n"
    "move.l (%sp)+,%a0");
#endif

        if (count < NR_PARTICLES-1) FillParticlesArray();
        ProcessParticlesArray();

#ifndef USE_TI89
        // check if screen is completely black ...
        if (   *(unsigned long long *)(LCD_MEM + 30)==-1ULL
            && *(unsigned long long *)(LCD_MEM + 38)==-1ULL
            && *(unsigned long long *)(LCD_MEM + 46)==-1ULL
            && *(unsigned long long *)(LCD_MEM + 52)==-1ULL) goto begin;
#else
        if (   *(unsigned long long *)(LCD_MEM + 30)==-1ULL
            && *(unsigned long long *)(LCD_MEM + 38)==-1ULL
            && *(unsigned long *)(LCD_MEM + 46)==-1UL)       goto begin;
#endif

        // is there a key waiting?
        if (!OSdequeue(&key,kbq)) {
            if (key == KEY_CLEAR) goto begin;
            else break;
        }
    }
    OSTimerRestart(APD_TIMER);

    //---------------------------------------------
    // cleanup and bye =8-)
    //---------------------------------------------
    free(positions);
    LCD_restore(lcd);
    GKeyFlush();
}
